/* 
 * Header File della classe Kernels
 */

#ifndef KERNELS_H
#define KERNELS_H

#include <../Eigen/Eigen/Core>
#include <cmath>

class Kernel
{
public:
	virtual ~Kernel() {}
	virtual double Eval(const Eigen::VectorXd& x1, const Eigen::VectorXd& x2) const = 0;
	virtual double Eval(const Eigen::VectorXd& x) const = 0;
};

class KernelLineare : public Kernel
{
public:
	inline double Eval(const Eigen::VectorXd& x1, const Eigen::VectorXd& x2) const
	{
		return x1.dot(x2);
	}
	
	inline double Eval(const Eigen::VectorXd& x) const
	{
		return x.squaredNorm();
	}
};

class KernelGaussiano : public Kernel
{
public:
	KernelGaussiano(double sigma) : Sigma(sigma) {}
	inline double Eval(const Eigen::VectorXd& x1, const Eigen::VectorXd& x2) const
	{
		return exp(-Sigma *(x1-x2).squaredNorm());
	}
	
	inline double Eval(const Eigen::VectorXd& x) const
	{
		return 1.0;
	}

private:
	double Sigma;//Sigma del Kernel Gaussiano
};

class KernelIntersection : public Kernel
{
public:
	inline double Eval(const Eigen::VectorXd& x1, const Eigen::VectorXd& x2) const
	{
		return x1.array().min(x2.array()).sum();
	}
	
	inline double Eval(const Eigen::VectorXd& x) const
	{
		return x.sum();
	}
};

class KernelChi2 : public Kernel
{
public:
	inline double Eval(const Eigen::VectorXd& x1, const Eigen::VectorXd& x2) const
	{
		double result = 0.0;
		for (int i = 0; i < x1.size(); ++i)
		{
			double a = x1[i];
			double b = x2[i];
			result += (a-b)*(a-b)/(0.5*(a+b)+1e-8);
		}
		return 1.0 - result;
	}
	
	inline double Eval(const Eigen::VectorXd& x) const
	{
		return 1.0;
	}
};

class MultiKernel : public Kernel
{
public:
	MultiKernel(const std::vector<Kernel*>& kernels, const std::vector<int>& featureCounts) :
		N(kernels.size()),
		Norm(1.0/kernels.size()),
		Miei_Kernel(kernels),
		Counts(featureCounts)
	{
	}
	
	inline double Eval(const Eigen::VectorXd& x1, const Eigen::VectorXd& x2) const
	{
		double sum = 0.0;
		int start = 0;
		for (int i = 0; i < N; ++i)
		{
			int c = Counts[i];
			sum += Norm * Miei_Kernel[i]->Eval(x1.segment(start, c), x2.segment(start, c));
			start += c;
		}
		return sum;	
	}
	
	inline double Eval(const Eigen::VectorXd& x) const
	{
		double sum = 0.0;
		int start = 0;
		for (int i = 0; i < N; ++i)
		{
			int c = Counts[i];
			sum += Norm * Miei_Kernel[i]->Eval(x.segment(start, c));
			start += c;
		}
		return sum;	
	}
	
private:
	int N;
	double Norm;
	std::vector<Kernel*> Miei_Kernel;
	std::vector<int> Counts;
	
};

#endif
