/* 
 * File cpp per la classe ClassificatoreLaRank
 */

#include "OnlineLearning.h"
#include "FileConfigurazione.h"
#include "Features.h"
#include "Kernels.h"
#include "UtiliGrafica/UtiliGrafica.h"
#include <opencv2/highgui/highgui.hpp>

static const int kTileSize = 30;
using namespace cv;

using namespace std;
using namespace Eigen;

static const int kMaxSVs = 2000; // Usato solo quando non è specificato alcun budget


OnlineLearning::OnlineLearning(const FileConfigurazione & conf, const Features& features, const Kernel& kernel) :
	Mia_Configurazione(conf),
	Mie_Features(features),
	Mio_Kernel(kernel),
	m_C(conf.svmC)
{
	int N = conf.svmBudgetSize > 0 ? conf.svmBudgetSize+2 : kMaxSVs;
	m_K = MatrixXd::Zero(N, N);
	ImmagineDebug = Mat(800, 600, CV_8UC3);
}

OnlineLearning::~OnlineLearning()
{
}

double OnlineLearning::Valutazione(const Eigen::VectorXd &x, const FloatRect &y) const
{
	double f = 0.0;
	for (int i = 0; i < (int) Vettori_Supporto.size(); ++i)
	{
		const SupportVector& sv = *Vettori_Supporto[i];
		f += sv.b* Mio_Kernel.Eval(x, sv.x->x[sv.y]);
	}
	return f;
}

void OnlineLearning::Eval(const MultiCampione & sample, std::vector<double>& results)
{
	const FloatRect& centre(sample.GetRects()[0]);
	vector<VectorXd> fvs;
	const_cast<Features&>(Mie_Features).Eval(sample, fvs);
	results.resize(fvs.size());
	for (int i = 0; i < (int)fvs.size(); ++i)
	{
		// express y in coord frame of centre sample
		FloatRect y(sample.GetRects()[i]);
		y.Translate(-centre.XMin(), -centre.YMin());
		results[i] = Valutazione(fvs[i], y);
	}
}

void OnlineLearning::Aggiornamento(const MultiCampione &campione, int y)
{
	// add new support pattern
	SupportPattern* sp = new SupportPattern;
	const vector<FloatRect>& rects = campione.GetRects();
	FloatRect centre = rects[y];
	for (int i = 0; i < (int)rects.size(); ++i)
	{
		// express r in coord frame of centre campione
		FloatRect r = rects[i];
		r.Translate(-centre.XMin(), -centre.YMin());
		sp->yv.push_back(r);
		if (!Mia_Configurazione.quietMode && Mia_Configurazione.debugMode)
		{
			// Memorizza una miniatura per ogni campione
			Mat im(kTileSize, kTileSize, CV_8UC1);
			IntRect rect = rects[i];
			cv::Rect roi(rect.XMin(), rect.YMin(), rect.Width(), rect.Height());
			cv::resize(campione.GetImage().GetImmagine(0)(roi), im, im.size());
			sp->images.push_back(im);
		}
	}
	// Valuta le features per ogni campione
	sp->x.resize(rects.size());
	const_cast<Features&>(Mie_Features).Eval(campione, sp->x);
	sp->y = y;
	sp->refCount = 0;
	Pattern_Supporto.push_back(sp);

	ProcessNew((int) Pattern_Supporto.size()-1);
	BudgetMaintenance();
	
	for (int i = 0; i < 10; ++i)
	{
		Reprocess();
		BudgetMaintenance();
	}
}

void OnlineLearning::BudgetMaintenance()
{
	if (Mia_Configurazione.svmBudgetSize > 0)
	{
		while ((int) Vettori_Supporto.size() > Mia_Configurazione.svmBudgetSize)
		{
			BudgetMaintenanceRemove();
		}
	}
}

void OnlineLearning::Reprocess()
{
	ProcessOld();
	for (int i = 0; i < 10; ++i)
	{
		Ottimizza();
	}
}

double OnlineLearning::ComputeDual() const
{
	double d = 0.0;
	for (int i = 0; i < (int) Vettori_Supporto.size(); ++i)
	{
		const SupportVector* sv = Vettori_Supporto[i];
		d -= sv->b*Loss(sv->x->yv[sv->y], sv->x->yv[sv->x->y]);
		for (int j = 0; j < (int) Vettori_Supporto.size(); ++j)
		{
			d -= 0.5*sv->b* Vettori_Supporto[j]->b*m_K(i,j);
		}
	}
	return d;
}

void OnlineLearning::SMOStep(int ipos, int ineg)
{
	if (ipos == ineg) return;

	SupportVector* svp = Vettori_Supporto[ipos];
	SupportVector* svn = Vettori_Supporto[ineg];
	assert(svp->x == svn->x);
	SupportPattern* sp = svp->x;

#if VERBOSE
	cout << "SMO: gpos:" << svp->g << " gneg:" << svn->g << endl;
#endif	
	if ((svp->g - svn->g) < 1e-5)
	{
#if VERBOSE
		cout << "SMO: skipping" << endl;
#endif		
	}
	else
	{
		double kii = m_K(ipos, ipos) + m_K(ineg, ineg) - 2*m_K(ipos, ineg);
		double lu = (svp->g-svn->g)/kii;
		// no need to clamp against 0 since we'd have skipped in that case
		double l = min(lu, m_C*(int)(svp->y == sp->y) - svp->b);

		svp->b += l;
		svn->b -= l;

		// update gradients
		for (int i = 0; i < (int) Vettori_Supporto.size(); ++i)
		{
			SupportVector* svi = Vettori_Supporto[i];
			svi->g -= l*(m_K(i, ipos) - m_K(i, ineg));
		}
#if VERBOSE
		cout << "SMO: " << ipos << "," << ineg << " -- " << svp->b << "," << svn->b << " (" << l << ")" << endl;
#endif		
	}
	
	// check if we should remove either sv now
	
	if (fabs(svp->b) < 1e-8)
	{
		RimuoviSupportVector(ipos);
		if (ineg == (int) Vettori_Supporto.size())
		{
			// ineg and ipos will have been swapped during sv removal
			ineg = ipos;
		}
	}

	if (fabs(svn->b) < 1e-8)
	{
		RimuoviSupportVector(ineg);
	}
}

pair<int, double> OnlineLearning::MinGradiente(int ind)
{
	const SupportPattern* sp = Pattern_Supporto[ind];
	pair<int, double> minGrad(-1, DBL_MAX);
	for (int i = 0; i < (int)sp->yv.size(); ++i)
	{
		double grad = -Loss(sp->yv[i], sp->yv[sp->y]) - Valutazione(sp->x[i], sp->yv[i]);
		if (grad < minGrad.second)
		{
			minGrad.first = i;
			minGrad.second = grad;
		}
	}
	return minGrad;
}

void OnlineLearning::ProcessNew(int ind)
{
	// gradient is -f(x,y) since loss=0
	int ip = AggiungiSupportVector(
			Pattern_Supporto[ind], Pattern_Supporto[ind]->y, -Valutazione(Pattern_Supporto[ind]->x[Pattern_Supporto[ind]->y],
																		   Pattern_Supporto[ind]->yv[Pattern_Supporto[ind]->y]));

	pair<int, double> minGrad = MinGradiente(ind);
	int in = AggiungiSupportVector(Pattern_Supporto[ind], minGrad.first, minGrad.second);

	SMOStep(ip, in);
}

void OnlineLearning::ProcessOld()
{
	if (Pattern_Supporto.size() == 0) return;

	// choose pattern to process
	int ind = rand() % Pattern_Supporto.size();

	// find existing sv with largest grad and nonzero beta
	int ip = -1;
	double maxGrad = -DBL_MAX;
	for (int i = 0; i < (int) Vettori_Supporto.size(); ++i)
	{
		if (Vettori_Supporto[i]->x != Pattern_Supporto[ind]) continue;

		const SupportVector* svi = Vettori_Supporto[i];
		if (svi->g > maxGrad && svi->b < m_C*(int)(svi->y == Pattern_Supporto[ind]->y))
		{
			ip = i;
			maxGrad = svi->g;
		}
	}
	assert(ip != -1);
	if (ip == -1) return;

	// find potentially new sv with smallest grad
	pair<int, double> minGrad = MinGradiente(ind);
	int in = -1;
	for (int i = 0; i < (int) Vettori_Supporto.size(); ++i)
	{
		if (Vettori_Supporto[i]->x != Pattern_Supporto[ind]) continue;

		if (Vettori_Supporto[i]->y == minGrad.first)
		{
			in = i;
			break;
		}
	}
	if (in == -1)
	{
		// add new sv
		in = AggiungiSupportVector(Pattern_Supporto[ind], minGrad.first, minGrad.second);
	}

	SMOStep(ip, in);
}

void OnlineLearning::Ottimizza()
{
	if (Pattern_Supporto.size() == 0) return;
	
	// choose pattern to optimize
	int ind = rand() % Pattern_Supporto.size();

	int ip = -1;
	int in = -1;
	double maxGrad = -DBL_MAX;
	double minGrad = DBL_MAX;
	for (int i = 0; i < (int) Vettori_Supporto.size(); ++i)
	{
		if (Vettori_Supporto[i]->x != Pattern_Supporto[ind]) continue;

		const SupportVector* svi = Vettori_Supporto[i];
		if (svi->g > maxGrad && svi->b < m_C*(int)(svi->y == Pattern_Supporto[ind]->y))
		{
			ip = i;
			maxGrad = svi->g;
		}
		if (svi->g < minGrad)
		{
			in = i;
			minGrad = svi->g;
		}
	}
	assert(ip != -1 && in != -1);
	if (ip == -1 || in == -1)
	{
		// this shouldn't happen
		cout << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
		return;
	}

	SMOStep(ip, in);
}

int OnlineLearning::AggiungiSupportVector(SupportPattern *x, int y, double g)
{
	SupportVector* sv = new SupportVector;
	sv->b = 0.0;
	sv->x = x;
	sv->y = y;
	sv->g = g;

	int ind = (int) Vettori_Supporto.size();
	Vettori_Supporto.push_back(sv);
	x->refCount++;

#if VERBOSE
	cout << "Adding SV: " << ind << endl;
#endif

	// Aggiorno la Kernel Matrix
	for (int i = 0; i < ind; ++i)
	{
		m_K(i,ind) = Mio_Kernel.Eval(Vettori_Supporto[i]->x->x[Vettori_Supporto[i]->y], x->x[y]);
		m_K(ind,i) = m_K(i,ind);
	}
	m_K(ind,ind) = Mio_Kernel.Eval(x->x[y]);

	return ind;
}

void OnlineLearning::ScambiaSupportVectors(int ind1, int ind2)
{
	SupportVector* tmp = Vettori_Supporto[ind1];
	Vettori_Supporto[ind1] = Vettori_Supporto[ind2];
	Vettori_Supporto[ind2] = tmp;
	
	VectorXd row1 = m_K.row(ind1);
	m_K.row(ind1) = m_K.row(ind2);
	m_K.row(ind2) = row1;
	
	VectorXd col1 = m_K.col(ind1);
	m_K.col(ind1) = m_K.col(ind2);
	m_K.col(ind2) = col1;
}

void OnlineLearning::RimuoviSupportVector(int ind)
{
#if VERBOSE
	cout << "Removing SV: " << ind << endl;
#endif	

	Vettori_Supporto[ind]->x->refCount--;
	if (Vettori_Supporto[ind]->x->refCount == 0)
	{
		// also remove the support pattern
		for (int i = 0; i < (int) Pattern_Supporto.size(); ++i)
		{
			if (Pattern_Supporto[i] == Vettori_Supporto[ind]->x)
			{
				delete Pattern_Supporto[i];
				Pattern_Supporto.erase(Pattern_Supporto.begin()+i);
				break;
			}
		}
	}

	// make sure the support vector is at the back, this
	// lets us keep the kernel matrix cached and valid
	if (ind < (int) Vettori_Supporto.size()-1)
	{
		ScambiaSupportVectors(ind, (int) Vettori_Supporto.size() - 1);
		ind = (int) Vettori_Supporto.size()-1;
	}
	delete Vettori_Supporto[ind];
	Vettori_Supporto.pop_back();
}

void OnlineLearning::BudgetMaintenanceRemove()
{
	// find negative sv with smallest effect on discriminant function if removed
	double minVal = DBL_MAX;
	int in = -1;
	int ip = -1;
	for (int i = 0; i < (int) Vettori_Supporto.size(); ++i)
	{
		if (Vettori_Supporto[i]->b < 0.0)
		{
			// find corresponding positive sv
			int j = -1;
			for (int k = 0; k < (int) Vettori_Supporto.size(); ++k)
			{
				if (Vettori_Supporto[k]->b > 0.0 && Vettori_Supporto[k]->x == Vettori_Supporto[i]->x)
				{
					j = k;
					break;
				}
			}
			double val = Vettori_Supporto[i]->b* Vettori_Supporto[i]->b*(m_K(i,i) + m_K(j,j) - 2.0*m_K(i,j));
			if (val < minVal)
			{
				minVal = val;
				in = i;
				ip = j;
			}
		}
	}

	// adjust weight of positive sv to compensate for removal of negative
	Vettori_Supporto[ip]->b += Vettori_Supporto[in]->b;

	// remove negative sv
	RimuoviSupportVector(in);
	if (ip == (int) Vettori_Supporto.size())
	{
		// ip and in will have been swapped during support vector removal
		ip = in;
	}
	
	if (Vettori_Supporto[ip]->b < 1e-8)
	{
		// also remove positive sv
		RimuoviSupportVector(ip);
	}

	// update gradients
	// TODO: this could be made cheaper by just adjusting incrementally rather than recomputing
	for (int i = 0; i < (int) Vettori_Supporto.size(); ++i)
	{
		SupportVector& svi = *Vettori_Supporto[i];
		svi.g = -Loss(svi.x->yv[svi.y],svi.x->yv[svi.x->y]) - Valutazione(svi.x->x[svi.y], svi.x->yv[svi.y]);
	}	
}

void OnlineLearning::Debug()
{
	cout << Pattern_Supporto.size() << "/" << Vettori_Supporto.size() << " support patterns/vectors" << endl;
	AggiornaImmagineDebug();
	imshow("learner", ImmagineDebug);
}

void OnlineLearning::AggiornaImmagineDebug()
{
	ImmagineDebug.setTo(0);
	
	int n = (int) Vettori_Supporto.size();
	
	if (n == 0) return;
	
	const int kCanvasSize = 600;
	int SizeGriglia = (int)sqrtf((float)(n-1)) + 1;
	int SizeTile = (int)((float)kCanvasSize/ SizeGriglia);
	
	if (SizeTile < 5)
	{
		cout << "too many support vectors to display" << endl;
		return;
	}
	
	Mat temp(SizeTile, SizeTile, CV_8UC1);
	int x = 0;
	int y = 0;
	int ind = 0;
	float vals[kMaxSVs];
	memset(vals, 0, sizeof(float)*n);
	int drawOrder[kMaxSVs];
	
	for (int set = 0; set < 2; ++set)
	{
		for (int i = 0; i < n; ++i)
		{
			if (((set == 0) ? 1 : -1)* Vettori_Supporto[i]->b < 0.0) continue;
			
			drawOrder[ind] = i;
			vals[ind] = (float) Vettori_Supporto[i]->b;
			++ind;
			
			Mat I = ImmagineDebug(cv::Rect(x, y, SizeTile, SizeTile));
			resize(Vettori_Supporto[i]->x->images[Vettori_Supporto[i]->y], temp, temp.size());
			cvtColor(temp, I, CV_GRAY2RGB);
			double w = 1.0;
			rectangle(I, Point(0, 0), Point(SizeTile -1, SizeTile -1), (Vettori_Supporto[i]->b > 0.0) ? CV_RGB(0, (uchar)(255*w), 0) : CV_RGB((uchar)(255*w), 0, 0), 3);
			x += SizeTile;
			if ((x+ SizeTile) > kCanvasSize)
			{
				y += SizeTile;
				x = 0;
			}
		}
	}
	
	const int SizekKernelPixel = 2;
	int SizeKernel = SizekKernelPixel *n;
	
	double kmin = m_K.minCoeff();
	double kmax = m_K.maxCoeff();
	
	if (SizeKernel < ImmagineDebug.cols && SizeKernel < ImmagineDebug.rows)
	{
		Mat K = ImmagineDebug(cv::Rect(ImmagineDebug.cols- SizeKernel, ImmagineDebug.rows- SizeKernel, SizeKernel,
									   SizeKernel));
		for (int i = 0; i < n; ++i)
		{
			for (int j = 0; j < n; ++j)
			{
				Mat Kij = K(cv::Rect(j* SizekKernelPixel, i* SizekKernelPixel, SizekKernelPixel, SizekKernelPixel));
				uchar v = (uchar)(255*(m_K(drawOrder[i], drawOrder[j])-kmin)/(kmax-kmin));
				Kij.setTo(Scalar(v, v, v));
			}
		}
	}
	else
	{
		SizeKernel = 0;
	}
	
	
	Mat I = ImmagineDebug(cv::Rect(0, ImmagineDebug.rows - 200, ImmagineDebug.cols- SizeKernel, 200));
	I.setTo(Scalar(255,255,255));
	IplImage II = I;
	setGraphColor(0);
	drawFloatGraph(vals, n, &II, 0.f, 0.f, I.cols, I.rows);
}
