/* 
 * File cpp del Tracker
 */

#include "Tracker.h"
#include "FileConfigurazione.h"
#include "Immagine.h"
#include "Campionatore.h"
#include "CampioneSingolo.h"

/* Features */
#include "FeaturesHaarMultiple.h"
#include "RawFeatures.h"
#include "FeaturesIstogramma.h"
#include "FeaturesMultiple.h"

#include "Kernels.h"
#include "OnlineLearning.h"

#include <opencv/highgui.h>

using namespace cv;
using namespace std;
using namespace Eigen;

Tracker::Tracker(const FileConfigurazione & conf) :
	Mia_Configurazione(conf),
	Mio_Inizializzato(false),
	Mio_PLearner(0),
	Mia_ImmagineDebug(2*conf.searchRadius+1, 2*conf.searchRadius+1, CV_32FC1),
	ServeImmagineIntegrale(false)
{
	Reset();
}

Tracker::~Tracker()
{
	delete Mio_PLearner;
	for (int i = 0; i < (int) Mie_Features.size(); ++i)
	{
		delete Mie_Features[i];
		delete m_kernels[i];
	}
}

void Tracker::Reset()
{
	Mio_Inizializzato = false;
	Mia_ImmagineDebug.setTo(0);
	if (Mio_PLearner) delete Mio_PLearner;
	for (int i = 0; i < (int) Mie_Features.size(); ++i)
	{
		delete Mie_Features[i];
		delete m_kernels[i];
	}
	Mie_Features.clear();
	m_kernels.clear();
	
	ServeImmagineIntegrale = false;
	m_needsIntegralHist = false;
	
	int numFeatures = Mia_Configurazione.TipologiaFeatures.size();
	vector<int> featureCounts;
	for (int i = 0; i < numFeatures; ++i)
	{
		//cout<< Mia_Configurazione.TipologiaFeatures[i].feature << endl;
		switch (Mia_Configurazione.TipologiaFeatures[i].feature)
		{

		case FileConfigurazione::kFeatureTipoHaar:
			Mie_Features.push_back(new FeaturesHaarMultiple(Mia_Configurazione));
			ServeImmagineIntegrale = true;
			break;			
		case FileConfigurazione::kFeatureTipoRaw:
			Mie_Features.push_back(new RawFeatures(Mia_Configurazione));
		    //std::cout<<"Calcolo le ASIFT"<<std::endl;
			//Mie_Features.push_back(new OrbFeatures(Mia_Configurazione));
			break;
		case FileConfigurazione::kFeatureTipoIstogramma:
			Mie_Features.push_back(new FeaturesIstogramma(Mia_Configurazione));
			m_needsIntegralHist = true;
			break;
		}
		featureCounts.push_back(Mie_Features.back()->GetCount());
		/*
		 *  Imposto il tipo di Kernel utilizzato
		 */
		switch (Mia_Configurazione.TipologiaFeatures[i].kernel)
		{
		case FileConfigurazione::kKernelTipoLineare:
			m_kernels.push_back(new KernelLineare());
			break;
		case FileConfigurazione::kKernelTipoGaussiano:
			m_kernels.push_back(new KernelGaussiano(Mia_Configurazione.TipologiaFeatures[i].params[0]));
			break;
		case FileConfigurazione::kKernelTipoIntersection:
			m_kernels.push_back(new KernelIntersection());
			break;
		case FileConfigurazione::kKernelTipoChi2:
			m_kernels.push_back(new KernelChi2());
			break;
		}
	}
	
	if (numFeatures > 1)
	{
		FeaturesMultiple * f = new FeaturesMultiple(Mie_Features);
		Mie_Features.push_back(f);
		
		MultiKernel* k = new MultiKernel(m_kernels, featureCounts);
		m_kernels.push_back(k);		
	}
	
	Mio_PLearner = new OnlineLearning(Mia_Configurazione, *Mie_Features.back(), *m_kernels.back());
}
	

void Tracker::Initialise(const cv::Mat& frame, FloatRect bb)
{
	m_bb = IntRect(bb);
	Immagine image(frame, ServeImmagineIntegrale, m_needsIntegralHist);
	for (int i = 0; i < 1; ++i)
	{
		AggiornaLearner(image);
	}
	Mio_Inizializzato = true;
}

void Tracker::Track(const cv::Mat& frame)
{
	assert(Mio_Inizializzato);
	
	Immagine image(frame, ServeImmagineIntegrale, m_needsIntegralHist);
	
	vector<FloatRect> rects = Campionatore::CampioniPixel(m_bb, Mia_Configurazione.searchRadius);
	//vector<FloatRect> rects = Campionatore::CampioniCirconferenza(m_bb,2*Mia_Configurazione.searchRadius,5,16);

	vector<FloatRect> keptRects;
	keptRects.reserve(rects.size());
	for (int i = 0; i < (int)rects.size(); ++i)
	{
		if (!rects[i].Interno(image.GetRect())) continue;
		keptRects.push_back(rects[i]);
	}
	
	MultiCampione sample(image, keptRects);
	
	vector<double> scores;
	Mio_PLearner->Eval(sample, scores);
	
	double bestScore = -DBL_MAX;
	int bestInd = -1;
	for (int i = 0; i < (int)keptRects.size(); ++i)
	{		
		if (scores[i] > bestScore)
		{
			bestScore = scores[i];
			bestInd = i;
		}
	}

	AggiornaImmagineDebug(keptRects, m_bb, scores);
	
	if (bestInd != -1)
	{
		m_bb = keptRects[bestInd];
		AggiornaLearner(image);
#if VERBOSE		
		cout << "track score: " << bestScore << endl;
#endif
	}
}

void Tracker::AggiornaImmagineDebug(const vector<FloatRect> &samples, const FloatRect &centre,
									const vector<double> &scores)
{
	double mn = VectorXd::Map(&scores[0], scores.size()).minCoeff();
	double mx = VectorXd::Map(&scores[0], scores.size()).maxCoeff();
	Mia_ImmagineDebug.setTo(0);
	for (int i = 0; i < (int)samples.size(); ++i)
	{
		int x = (int)(samples[i].XMin() - centre.XMin());
		int y = (int)(samples[i].YMin() - centre.YMin());
		Mia_ImmagineDebug.at<float>(Mia_Configurazione.searchRadius+y, Mia_Configurazione.searchRadius+x) = (float)((scores[i]-mn)/(mx-mn));
	}
}

void Tracker::Debug()
{
	imshow("tracker", Mia_ImmagineDebug);
	Mio_PLearner->Debug();
}

void Tracker::AggiornaLearner(const Immagine &image)
{
	// note these return the centre sample at index 0
	vector<FloatRect> rects = Campionatore::CampioniCirconferenza(m_bb, 2 * Mia_Configurazione.searchRadius, 5, 16);
	//vector<FloatRect> rects = Campionatore::CampioniPixel(m_bb, 2*Mia_Configurazione.searchRadius, true);
	
	vector<FloatRect> keptRects;
	keptRects.push_back(rects[0]); // the true sample
	for (int i = 1; i < (int)rects.size(); ++i)
	{
		if (!rects[i].Interno(image.GetRect())) continue;
		keptRects.push_back(rects[i]);
	}
		
#if VERBOSE		
	cout << keptRects.size() << " samples" << endl;
#endif
		
	MultiCampione sample(image, keptRects);
	Mio_PLearner->Aggiornamento(sample, 0);
}

double Tracker::GetFeaturesVector(int index){
	return (Mie_Features[Mie_Features.size()-1]->m_featVec[index]);
}
